#include <cstdlib>
using std::exit;

#include <cstddef>
using std::size_t;

#include <cstring>
using std::strlen;

#include <sstream>
using std::stringstream;

#include <cassert>

#ifdef USE_OPENGL
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#endif // USE_OPENGL

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
using std::fixed;

#include <fstream>
using std::ofstream;

#include <iomanip>
using std::setw;

#include <cmath>
using std::atan2;
using std::cos;
using std::sin;
using std::sqrt;

#include <vector>
using std::vector;

#include <algorithm>
using std::min;

#include "particle.hpp"

vector<particle> gas; // gas = vector of particles
const double dt = 0.0005;
const double dt2 = dt*dt;
double global_time = 0.0; // a global time counter

#ifdef USE_OPENGL
const int delay = 10; // milliseconds

void glutPrint(float x, float y, const char* text, float r, float g, float b, float a) 
{ 
    if(!text || !strlen(text)) return; 
    bool blending = false; 
    if(glIsEnabled(GL_BLEND)) blending = true; 
    glEnable(GL_BLEND); 
    glColor4f(r,g,b,a); 
    glRasterPos2f(x,y); 
    while (*text) { 
        glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *text); 
        text++; 
    } 
    if(!blending) glDisable(GL_BLEND); 
}

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t n = 0; n < gas.size(); ++n)
	{
		const double offset = gas[n].draw();
		if (!gas[n].name.empty() and gas[n].name != "_")
			glutPrint(gas[n].r.x+offset,gas[n].r.y+offset,gas[n].name.c_str(), 1.0f, 1.0f, 1.0f, 0.5f);
	}
	stringstream s;
	s.precision(1);
	s << "t = " << fixed << global_time;
	glutPrint(-0.85f, -0.85f,s.str().c_str(), 1.0f, 1.0f, 1.0f, 0.5f);
	glFlush();	
	glutSwapBuffers();
}

void keyboard(unsigned char key, int x, int y)
{
	if (key == 'q')
		exit(0);
}
#endif // USE_OPENGL

inline double angle(const particle &p1, const particle &p2)
{
	const double dx = p2.r.x - p1.r.x;
	const double dy = p2.r.y - p1.r.y;
	return atan2(dy,dx);
}

inline double distance2(const particle &p1, const particle &p2)
{
	const double dx = p2.r.x - p1.r.x;
	const double dy = p2.r.y - p1.r.y;
	return dx*dx + dy*dy;
}

vec2 recompute_acceleration(size_t n)
{
	double sumx = 0.0;
	double sumy = 0.0;
	for (size_t m = 0; m < gas.size(); ++m)
		if (n != m)
		{
			const double r2 = distance2(gas[n],gas[m]);
			const double theta = angle(gas[n],gas[m]);
			sumx += gas[m].mass*cos(theta)/r2;
			sumy += gas[m].mass*sin(theta)/r2;
		}
	return vec2(sumx,sumy);
}

void initialize_gas(void)
{
	gas.push_back( particle( /*name*/      "A",
							 /*mass*/      1.0,
	                         /*position*/  vec2(-0.4,0.05),
	                         /*velocity*/  vec2(0.094285714285714,12.6568303571429) ) );
	gas.push_back( particle( /*name*/      "B",
							 /*mass*/      2.0,
	                         /*position*/  vec2(0.7,-0.1),
	                         /*velocity*/  vec2(-2.54571428571428,-7.75941964285714) ) );
	gas.push_back( particle( /*name*/      "C",
							 /*mass*/      3.0,
	                         /*position*/  vec2(0.0,0.3),
	                         /*velocity*/  vec2(15.0942857142857,0.056830357142857) ) );
	gas.push_back( particle( /*name*/      "D",
							 /*mass*/      50.0,
	                         /*position*/  vec2(0.0316,0.0877),
	                         /*velocity*/  vec2(-0.805714285714286,0.053830357142857) ) );
	
	for (size_t n = 0; n < gas.size(); ++n)
		gas[n].a = recompute_acceleration(n);

	cout << "Simulating " << gas.size() << " particles..." << endl;
}

void update_gas(void)
{
	// self-starting Verlet algorithm
	// x(n+1) = x(n) + v(n)dt + 0.5*a(n)dt^2
	// v(n+1) = v(n) + 0.5*(a(n+1)+a(n))*dt
	
	// a(n) stored in gas[n].a from the previous time step
	for (size_t n = 0; n < gas.size(); ++n)
		gas[n].r += gas[n].v*dt + gas[n].a*(0.5*dt2); // x(n) -> x(n+1)
	for (size_t n = 0; n < gas.size(); ++n)
	{
		const vec2 a_old = gas[n].a;
		const vec2 a_new = gas[n].a = recompute_acceleration(n); // a(n+1)
		gas[n].v += (a_new+a_old)*0.5*dt; // v(n+1) = v(n) + 0.5*(a(n+1)+a(n))*dt
		gas[n].t += dt;
	}
	
	global_time += dt; // increment the time counter
}

#ifdef USE_OPENGL
void animate(int)
{
	update_gas();
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

int main(int argc, char **argv)
{
	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (640, 640);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("A three-planet solar system");
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);	
	glutTimerFunc(delay,animate,0);
	initialize_gas();
	glutMainLoop();
}
#else
int main(int argc, char **argv)
{
	initialize_gas();
	const size_t N = gas.size();
	ofstream *fout = new ofstream[N];
	for (size_t k = 0; k < N; ++k)
	{
		string filename = "trajectory-" + gas[k].name + ".dat";
		fout[k].open(filename.c_str());
	}
	while (gas[0].t < 10.0)
	{
		update_gas();
		for (size_t k = 0; k < N; ++k)
			fout[k] << gas[k] << endl;
	}	
	for (size_t k = 0; k < N; ++k)
		fout[k].close();
	delete [] fout;	
	return 0;
}
#endif // USE_OPENGL
